# Copyright (c) 2004-2023 Yury V. Vishnevskiy yu.v.vishnevskiy@gmail.com
# All Rights Reserved.
# This file is subject to the terms and conditions defined in the
# 'license.txt' file, which is part of this package.

param (
 [Parameter(Mandatory=$false)][string]$unexdir
)

$EXITCODE = 0

$UDIR = $unexdir

$TFILECUR = ''

function Check-Reference($logfile, $reffile, $errfile) {
  $failed=0
  $skiplines=0
  $loglines = Get-Content $logfile
  $reflines = Get-Content $reffile
  foreach ($refline in $reflines) {
    $ssret = ($loglines | Select-Object -Skip $skiplines) | Select-String -Pattern $refline -SimpleMatch -CaseSensitive -List | Select-object -First 1
    if ($ssret) {
      # all good
      $skiplines += $ssret.LineNumber
    }
    else {
      $failed++
      Add-Content -Path $errfile -Value "NoRefStr(fromline=$($skiplines+1)): $refline"
    }
  }
  $failed
}


try {
  #(Get-Variable PSVersionTable -ValueOnly).PSVersion
  if (($global:UDIR -eq $null) -or ($global:UDIR -eq "")) {
    $global:UDIR = Split-Path -Parent -Path $MyInvocation.MyCommand.Definition
  }
  $udir = $global:UDIR
  
  $unex = $udir + "\unex.exe"
  $testdir = $udir + "\tests\public\inp"
  
  $timeoutSeconds = 120
  
  $elapsedTime = [System.Diagnostics.Stopwatch]::StartNew()
  
  # Clean up after last run
  # Deleting all files except *.inp and *.ref
  # Wrap in array @( ) to prevent iterating in forearch over null in PowerShell <= 2.
  $ofiles = @( Get-ChildItem $testdir -Exclude "*.inp","*.ref" -Recurse | Where-Object { !$_.PSIsContainer } | Select -ExpandProperty FullName )
  foreach ($ofile in $ofiles) { Remove-Item -Force $ofile }
  # Delete empty subdirs
  do {
    $ddirs = @(Get-ChildItem $testdir -Recurse | Where-Object { $_.PSIsContainer } | Where { @(Get-ChildItem $_.FullName -Force).Count -eq 0 } | Select -ExpandProperty FullName)
    $ddirs | Foreach-Object { Remove-Item $_ }
  } while ($ddirs.Count -gt 0)
  
  # Wrap in array @( ) to prevent iterating in forearch over null in PowerShell <= 2.
  $tfiles = @( Get-ChildItem $testdir -Filter "*.inp" -Recurse | Select -ExpandProperty FullName )

  $ttotal = $tfiles.Count
  $tcurrent = 0
  $tpassed  = 0
  $tpassed_ref = 0
  $tfailed  = 0
  $tfailed_ref = 0
  $tskippd  = 0
  $ttimeout = 0

  foreach ($tfile in $tfiles) {
    $tcurrent++
    $global:TFILECUR = $tfile.Replace($udir+"\", '')
    $curmsg = "$tcurrent of $ttotal  $global:TFILECUR ... "
    
    $tdir = [IO.Path]::GetDirectoryName($tfile)
    $tifname = [IO.Path]::GetFileName($tfile)
    $tofname = $tdir + "\" + [IO.Path]::GetFileNameWithoutExtension($tfile) + ".log" # with absolute path
    $tefname = $tdir + "\" + [IO.Path]::GetFileNameWithoutExtension($tfile) + ".err" # with absolute path
    $trfname = $tdir + "\" + [IO.Path]::GetFileNameWithoutExtension($tfile) + ".ref" # with absolute path

    # Skip tests if need
    if ($tifname -match "Ih_C540") {
      Write-Output ($curmsg + "Skipped")
      $tskippd++
      Continue
    }

    $code = {
      & $args[0] -d $args[1] $args[2] $args[3] > $null 2> $args[4]
      $LastExitCode
    }
    
    $job = Start-Job -ScriptBlock $code -ArgumentList @($unex,$tdir,$tifname,$tofname,$tefname)
    
    if (Wait-Job $job -Timeout $timeoutSeconds) {
      # Get here if the job is no more running
      if ($job.State -eq "Completed") {
	$lec = Receive-Job $job

        if ([int]$lec -eq 0) {
	  # If reference file exists, do checks
	  if (Test-Path -Path $trfname -PathType Leaf) {
	    $ret = Check-Reference $tofname $trfname $tefname
	    if ([int]$ret -eq 0) {
              Write-Output ($curmsg + "Successful")
              $tpassed++
              $tpassed_ref++
	    }
            else {
              Write-Output ($curmsg + "Failed")
              $tfailed++
              $tfailed_ref++
            }
	  }
          else {
            Write-Output ($curmsg + "Successful")
            $tpassed++
	  }
        }
        else {
          Write-Output ($curmsg + "Failed")
          $tfailed++
        }
      }
      else {
        Write-Output ($curmsg + "Failed")
        $tfailed++
      }
    }
    else {
      Remove-Job -Force $job
      Write-Output ($curmsg + "Timeout")
      $ttimeout++
    }
    
    # Remove empty *.err file
    if ((Test-Path -Path $tefname -PathType Leaf) -And ((Get-Content $tefname) -eq $Null)) {
      Remove-Item -Force $tefname
    }
    
  }
}
catch {
  Write-Output "---------------------------------------------------------------------------"
  Write-Output "Error:"
  $_
  Write-Output " "
  Write-Output "Current test: $global:TFILECUR"
  Write-Output "---------------------------------------------------------------------------"
  $EXITCODE = 1
}
finally {
  # Code to run on exit
  Write-Output "Skipped: $tskippd"
  Write-Output "Passed:  $tpassed ($tpassed_ref with external reference)"
  Write-Output "Failed:  $tfailed ($tfailed_ref with external reference)"
  Write-Output "Timeout: $ttimeout"
  if ([int]$tfailed -ne 0) { $EXITCODE = 1 }
  if ([int]$ttimeout -ne 0) { $EXITCODE = 1 }
  
  Write-Output "Total elapsed time: $($elapsedTime.Elapsed.ToString())"
}

exit $EXITCODE
